/*	Copyright: autocart (user name at the ahk and XYplorer forums) unless otherwise noted in the code
Version: 3.1
Date: 2020-04-24
Possible ideas for the future:

-) not only a black list (excludePaths) but also a white list, which would indicate that only the content of the white list gets redirected;
-) a switch, indicating if either the white list is considered first and the black list is only a sub of the white overwriting certain exceptions OR the other way round - the black list being considered first and the white being a sub of the black overwriting certain exceptions
-) making the excludePaths content (and a possible white list's content) respect wild cards, or even better, be defined in RegEx
-) Re-introduce the handling of existing WE windows at script start (in a better way and maybe handle also other message scenarios - not just "window created")
-) global hotkey(s) for pausing and un-pausing the script
-) improve the MsgBox popping in case the path could not be determined (maybe with debug-information that a user could then post in the forum for my knowledge) and make it easier for the end user to disable that MsgBox if he/she would want to not see it
-) 

General description:

Preamble:
Most of the code was written years ago. Looking at the code now, it does not look pretty at all. I feel a need to rewrite it from scratch but don't have the time nor the motivation for it at the moment. For now, I just changed the things, which seemed to me most important. In a later version I might rewrite the code - or not.

Purpose and usage of the script:
This script works in combination with the file explorer XYplorer (from now on called "XY") and Windows Explorer (from now on called "WE").
It was written for the following scenarios:

(1)	XY's setting "XY is default file manager" not getting respected:
	Despite XY's build-in optional functionality to act as the system's default file manager (XY menu Tools/Configuration/Other/Shell Integration), sometimes there still are WE windows opening instead of XY (e.g. from within certain 3rd-party-software or for other reasons).

(2)	List items not getting automatically selected in XY:
	At other times XY *is* sucessfully used as the system's default file manager to open a folder (if that setting is on) but any list itmes which would normally be selected in WE do *not* get selected in XY (e.g. when clicking in Firefox in the "downloads" window on the "open containing folder" button).

For such scenarios, this script, when running, reads the path shown in any WE window being opened, closes that WE window, opens a new tab in XY with that path *and* selects any list items in XY that where selected in WE before (and scrolling them into view), thus "redirecting" the WE window to XY.

The script will stay in the system tray until manually exited.

Hint:
For the script to work also in cases of scenario No. (2), of course, the XY setting to act as the system's default file manager must be turned *off*.

More details:

If XY is not running it will be started automatically if the path to XYplorer.exe is specified correctly. The path's default value is "C:\Program Files (x86)\XYplorer\XYplorer.exe". If the path needs to be changed, it can be specified in the variable $pathToXYplorer at the beginning of the code below. A tooltip close to the mouse cursor will indicate the starting process of XYplorer. If the XY-autostart should not work for some reason the script will display a message box and wait for the user to start XY manually.

WE windows, which show the content of a zip file, are not redirected to XY, because XY does not support this kind of view.

If one wants any other WE windows as well to *not* be redirected to XY (e.g. control panel stuff), then there are these options:
(1)	Control-key-escape:
	Hold the left or right control key pressed down while the WE window opens.
	Hint:
	With the script running, pressing Win+E would open a new WE window which normally would get redirected to XY.
	Pressing Control+Win+E, the new WE window would *not* get redirected to XY. (Of course the Control-key-escape option should work also if a WE window is opened by other means.)

(2)	Command line parameters:
	To permanently prevent certain WE windows from being redirected to XY, the script accepts command line parameters containing paths or keywords to be excluded from redirection (e.g. "Computer", "Libraries", "Pictures" ... or real whole paths without trailing backslash). For more suggestions for keywords see No. (3) below. Each path or keyword must be stated separately as its own command line parameter. The command line (or shortcut target line) could then, e.g., look like this:
	"path_to_autohotkey.exe" "path_to_script.ahk" "Computer" "Libraries" "Pictures" "C:\some kind of folder\another folder"
	or for a compiled version of the script:
	"path_to_script.exe" "Computer" "Libraries" "Pictures" "C:\some kind of folder\another folder"

(3)	Hard coded exclude-paths:
	Like No. (2) but not as command line parameters, one can also hard code the paths and keywords into the code (suggestions below), in which case the command line parameters would become unnecessary (but would still be allowed additionally). To do so, look below, where the code begins, for the line which starts with:
	; $excludePaths := "|All Control Panel Items|..."
	Remove the first ";"-character in that line and change the keywords and paths accordingly to your needs.

Known limitations:

-)	Whenever a WE window is "redirected" to XY, it will still open for a short moment in a normal way and immediately be closed again. For this script to work, I do not know of any way how to avoid that.
-)	This script has not been tested with WE add-ons, which add tabs to WE.
-)	This script has not been tested with the scenario of multiple XY windows.

For suggestions and ideas for improvement I am grateful.

Changes in version 3.0:

-)	Improved reliability:
	Before, the script sometimes missed WE windows or seemingly could not retrieve its path and then did not react as expected.
-)	Improved behaviour when paused:
	Now the script is respecting the script's paused state (see the script's system tray context menu) in that it stops redirecting when the script paused.
	Before it ignored the state of being paused and redirected anyway.
-)	Changed the behaviour of the script regarding already existing WE windows:
	In previous versions, when the title bar of already existing WE windows (which existed prior to the scripts start) changed (e.g. when changing a folder in such a window or when refreshing by hitting F5), then the script also used to redirect such WE windows to XY.
	Not anymore. Now, if a WE window already exist at the scripts start, the script leaves it untouched.
-)	Added a Control-key-escape option:
	If the left or right Control key is pressed while a WE window opens, then this particular WE window will *not* get redirected to XY.

Changes in version 3.1:

-)	Fixed the problem that the script would crash if XY was not active.
-)	The path to XY can now be specified in the variable $pathToXYplorer at the beginning of the code.
-)	In case of a XY autostart process, a tooltip will be diplayed close to the mouse cursor.
-)	In case the XY autostart process could not be started, a message box will pop.
*/


#NoEnv  ; Recommended for performance and compatibility with future AutoHotkey releases.
#SingleInstance force
#Persistent
; #Warn  ; Enable warnings to assist with detecting common errors.
SendMode Input  ; Recommended for new scripts due to its superior speed and reliability.
SetWorkingDir %A_ScriptDir%  ; Ensures a consistent starting directory.

global $pathToXYplorer
$pathToXYplorer := "C:\Program Files (x86)\XYplorer\XYplorer.exe"
global $excludePaths
$excludePaths := "|"
; $excludePaths := "|All Control Panel Items|C:\sample path|Systemsteuerung|Storage Spaces|Sync Center|System|Network and Sharing Center|Programs and Features|Recovery|RemoteApp and Desktop Connections|Security and Maintenance|Speech Recognition|Autoplay|Fonts|Backup and Restore (Windows 7)|Devices and Printers|Dolby Audio|Ease of Access Center|File History|BitLocker Drive Encryption|Color Management|Credential Manager|Device Manager|Action Center|Administrative Tools|Default Programs|Mouse|Power Options|Personalization|User Accounts|Windows 10 Firewall Control|Troubleshooting|Windows Defender Firewall|Windows Firewall|Work Folders|"

Loop, %0% 
{
	$excludePaths := $excludePaths . %A_Index% . "|"
}
StringReplace, $excludePaths, $excludePaths, ",, All
	
Gui +LastFound
$hWnd := WinExist()
DllCall( "RegisterShellHookWindow", UInt, $hWnd )
MsgNum := DllCall( "RegisterWindowMessage", Str, "SHELLHOOK" )
OnMessage( MsgNum, "ShellMessage" )
Return

_WinGetClass(WinTitle := "", WinText := "", ExcludeTitle := "", ExcludeText := "")
{
	WinGetClass, class, %WinTitle%, %WinText%, %ExcludeTitle%, %ExcludeText%
	return class
}

^#e::
OnMessage( MsgNum, "ShellMessage", 0 )
OnMessage( MsgNum, "ShellMessage_forCtrlWinE" )
Send, #e
return

ShellMessage_forCtrlWinE( wParam, lParam )
{
	Critical, On
	global MsgNum
	if (wParam = 1 and _WinGetClass("ahk_id " lParam) = "CabinetWClass")
	{
		OnMessage( MsgNum, "ShellMessage_forCtrlWinE", 0 )
		OnMessage( MsgNum, "ShellMessage" )
	}
}

ShellMessage( wParam, lParam )
{
	Critical, On
	if A_IsPaused
		return

/*wParam =
HSHELL_ACCESSIBILITYSTATE
11
The accessibility state has changed.

HSHELL_ACTIVATESHELLWINDOW
3
The shell should activate its main window.

HSHELL_APPCOMMAND
12
The user completed an input event (for example, pressed an application command button on the mouse or an application command key on the keyboard), and the application did not handle the WM_APPCOMMAND message generated by that input.
If the Shell procedure handles the WM_COMMAND message, it should not call CallNextHookEx. See the Return Value section for more information.

HSHELL_GETMINRECT
5
A window is being minimized or maximized. The system needs the coordinates of the minimized rectangle for the window.

HSHELL_LANGUAGE
8
Keyboard language was changed or a new keyboard layout was loaded.

HSHELL_REDRAW
6
The title of a window in the task bar has been redrawn.

HSHELL_TASKMAN
7
The user has selected the task list. A shell application that provides a task list should return TRUE to prevent Windows from starting its task list.

HSHELL_WINDOWACTIVATED
4
The activation has changed to a different top-level, unowned window.

HSHELL_WINDOWCREATED
1
A top-level, unowned window has been created. The window exists when the system calls this hook.

HSHELL_WINDOWDESTROYED
2
A top-level, unowned window is about to be destroyed. The window still exists when the system calls this hook.

HSHELL_WINDOWREPLACED
13
A top-level window is being replaced. The window exists when the system calls this hook. 

lParam differs in type according to the value of wParam received. For most of the wParam values, the lParam is a handle to a window that can be used as ahk_id %lParam% in AHK's Window commands.
*/

	global $pathToXYplorer
	global $excludePaths
	; static $hWndsToIgnore := "|"

	; if ( (wParam = 1 or wParam = 6) and WinActive("ahk_class CabinetWClass"))
	if (wParam = 1 and _WinGetClass("ahk_id " lParam) = "CabinetWClass")
	{
		; if (InStr($hWndsToIgnore, "|" lParam "|"))
		; 	return
		if (GetKeyState("Control", "P"))
		{
			; $hWndsToIgnore .= lParam "|"
			return
		}

		SetTitleMatchMode, 2
		IfWinNotExist, XYplorer ahk_class ThunderRT6FormDC
		{
			SplitPath, $pathToXYplorer, OutFileName, OutDir, OutExtension, OutNameNoExt, OutDrive
			; Run, XYPlorer.exe, OutDir, UseErrorLevel
			Run, %$pathToXYplorer%, OutDir, UseErrorLevel
			if (ErrorLevel = "ERROR")
			{
				while not WinExist("XYplorer ahk_class ThunderRT6FormDC")
				{
					MsgBox, 262145, %A_ScriptName%, XYplorer could not be started automatically. Please, start it manually.`n`nOnce XYplorer is started you can close this dialog in order to redirect the Windows Explorer window to XYplorer.`n`nPress Cancel in order to *not* redirect the Windows Explorer window to XYplorer this time.
					IfMsgBox, Cancel
						return
				}
			}
			else
			{
				; WinWait, XYplorer ahk_class ThunderRT6FormDC
				while not WinExist("XYplorer ahk_class ThunderRT6FormDC")
				{
					ToolTip, %A_ScriptName%:`n`nXYplorer is being started...
					Sleep, 20
				}
				ToolTip
			}
			IfWinNotExist, ahk_id %lParam%
				return			
		}

		$path := Explorer_GetPath(lParam)
		while ( $path = "ERROR" and A_Index < 10)
		{
			 Sleep 100
			 $path := Explorer_GetPath(lParam)
		}
		if ( $path = "ERROR" )
		{
			 MsgBox Could not retrieve the path of the explorer window.
			 return
		}
		$selected := Explorer_GetSelected(lParam)
		StringReplace, $selected, $selected, % Chr(10), |, 1
		StringReplace, $selected, $selected, ', '', 1
		If (StrLen($path) > 3 and StrLen($path) - InStr($path, ".zip", 0, 0) = 3)
			return
		If (InStr($excludePaths, "|" . $path . "|"))
		{
			return
		}
		
		SetTitleMatchMode, 2
		WinWait, XYplorer ahk_class ThunderRT6FormDC
		WinActivate
		WinClose, ahk_id %lParam%
		XY_send_message_noreturnvalue("::if(strpos(get('tabs'), '" . $path "') != -1){goto '" . $path "', 1;}else{tab('new', '" . $path . "');}")
		;~ XY_send_message_noreturnvalue("::goto '" . $path "', 1;")
		XY_send_message_noreturnvalue("::selectitems '" . $selected . "';#1012;") ;#1012 is XY-command-ID for Miscellaneous | Focus Functions | Move Focused Item Into View
	}
}


;*****************************************************************************************************************
;~ Code from others  (or based on):
;*****************************************************************************************************************
/*
	Library for getting info from a specific explorer window (if window handle not specified, the currently active
	window will be used).  Requires AHK_L or similar.  Works with the desktop.  Does not currently work with save
	dialogs and such.
	
	
	Explorer_GetSelected(hwnd="")   - paths of target window's selected items
	Explorer_GetAll(hwnd="")        - paths of all items in the target window's folder
	Explorer_GetPath(hwnd="")       - path of target window's folder
	
	example:
		F1::
			path := Explorer_GetPath()
			all := Explorer_GetAll()
			sel := Explorer_GetSelected()
			MsgBox % path
			MsgBox % all
			MsgBox % sel
		return
	
	Joshua A. Kinnison
	2011-04-27, 16:12
*/

Explorer_GetPath(hwnd="")
{
	if !(window := Explorer_GetWindow(hwnd))
		return ErrorLevel := "ERROR"
	if (window="desktop")
		return A_Desktop
	path := window.LocationURL
	if (path="")
		path:=window.LocationName
	path := RegExReplace(path, "ftp://.*@","ftp://")
	StringReplace, path, path, file:///
	StringReplace, path, path, /, \, All 
	
	; thanks to polyethene
	Loop
		If RegExMatch(path, "i)(?<=%)[\da-f]{1,2}", hex)
			StringReplace, path, path, `%%hex%, % Chr("0x" . hex), All
		Else Break
	return path
}

Explorer_GetAll(hwnd="")
{
	return Explorer_Get(hwnd)
}

Explorer_GetSelected(hwnd="")
{
	return Explorer_Get(hwnd,true)
}

Explorer_GetWindow(hwnd="")
{
	; thanks to jethrow for some pointers here
    WinGet, process, processName, % "ahk_id " hwnd := hwnd? hwnd:WinExist("A")
    WinGetClass class, ahk_id %hwnd%


	; MsgBox, %hwnd%, %process%, %class%
	; ; if not process
	; 	MsgBox, % WinExist("ahk_id " hwnd)
	; ; return
	
	if (process!="explorer.exe")
		return
	if (class ~= "(Cabinet|Explore)WClass")
	{
		for window in ComObjCreate("Shell.Application").Windows
			try if (window.hwnd==hwnd)
				return window
	}
	else if (class ~= "Progman|WorkerW") 
		return "desktop" ; desktop found
}

Explorer_Get(hwnd="",selection=false)
{
	if !(window := Explorer_GetWindow(hwnd))
		return ErrorLevel := "ERROR"
	if (window="desktop")
	{
		ControlGet, hwWindow, HWND,, SysListView321, ahk_class Progman
		if !hwWindow ; #D mode
			ControlGet, hwWindow, HWND,, SysListView321, A
		ControlGet, files, List, % ( selection ? "Selected":"") "Col1",,ahk_id %hwWindow%
		base := SubStr(A_Desktop,0,1)=="\" ? SubStr(A_Desktop,1,-1) : A_Desktop
		Loop, Parse, files, `n, `r
		{
			path := base "\" A_LoopField
			IfExist %path% ; ignore special icons like Computer (at least for now)
				ret .= path "`n"
		}
	}
	else
	{
		if selection
			collection := window.document.SelectedItems
		else
			collection := window.document.Folder.Items
		for item in collection
			ret .= item.path "`n"
	}
	return Trim(ret,"`n")
}

;*****************************************************************************************************************

;heavily based on
;http://www.xyplorer.com/xyfc/viewtopic.php?f=7&t=9233&p=89555&hilit=messenger#p89555

XY_send_message_noreturnvalue(MessagetoXYplorer)          ;Send message to XYplorer
{
  SenderHWND := A_ScriptHwnd + 0  ;Return this script's hidden hwdn id.  +0 to convert from Hex to Dec
  SetTitleMatchMode, 2
  HWND := WinExist("XYplorer ahk_class ThunderRT6FormDC")
  Size := StrLen(MessagetoXYplorer)
  If !(A_IsUnicode)
  {
  VarSetCapacity(Data, Size * 2, 0)
  StrPut(MessagetoXYplorer, &Data, Size, "UTF-16")
  }
  Else
  Data := MessagetoXYplorer

  VarSetCapacity(COPYDATA, A_PtrSize * 3, 0)
  NumPut(4194305, COPYDATA, 0, "Ptr")
  NumPut(Size * 2, COPYDATA, A_PtrSize, "UInt")
  NumPut(&Data, COPYDATA, A_PtrSize * 2, "Ptr")
  SendMessage, 0x4a, 0, &COPYDATA, , ahk_id %HWND% ;SendMessage waits for the target window to process the message, up until the timeout period expires.
}

global XY_eval_and_returnvalue_XYRETURN

XY_eval_and_returnvalue(expression)          ;Send message to XYplorer
{
  global XY_eval_and_returnvalue_XYRETURN
  SenderHWND := A_ScriptHwnd + 0  ;Return this script's hidden hwdn id.  +0 to convert from Hex to Dec
  MessagetoXYplorer := "::CopyData " SenderHWND ", " expression ", 0"    ;resolved to sth like this:   ::CopyData 7409230, <curitem>, 0   _OR like this:  ::CopyData 7409230, tab('get','count'), 0
  SetTitleMatchMode, 2
  HWND := WinExist("XYplorer ahk_class ThunderRT6FormDC")
  Size := StrLen(MessagetoXYplorer)
  If !(A_IsUnicode)
  {
  VarSetCapacity(Data, Size * 2, 0)
  StrPut(MessagetoXYplorer, &Data, Size, "UTF-16")
  }
  Else
  Data := MessagetoXYplorer

  VarSetCapacity(COPYDATA, A_PtrSize * 3, 0)
  NumPut(4194305, COPYDATA, 0, "Ptr")
  NumPut(Size * 2, COPYDATA, A_PtrSize, "UInt")
  NumPut(&Data, COPYDATA, A_PtrSize * 2, "Ptr")
  OnMessage(0x4a, "Function_Receive_WM_COPYDATA")  ; 0x4a is WM_COPYDATA. This onhold and wait for the WM_Copydata from XYplorer then execute Function_Receive_WM_COPYDATA(wParam, lParam) below
  SendMessage, 0x4a, 0, &COPYDATA, , ahk_id %HWND% ;SendMessage waits for the target window to process the message, up until the timeout period expires.
  OnMessage(0x4a, "")
  return XY_eval_and_returnvalue_XYRETURN
}

Function_Receive_WM_COPYDATA(wParam, lParam)
{
  global XY_eval_and_returnvalue_XYRETURN
  StringAddress := NumGet(lParam + 2*A_PtrSize) ;lParam+8 is the address of CopyDataStruct's lpData member.
  CopyOfData := StrGet(StringAddress)    ;May also specify CP0 (default) or UTF-8 or UTF-16:   StrGet(StringAddress, NumGet(lParam+A_PtrSize), "UTF-16")
  cbData := NumGet(lParam+A_PtrSize)/2  ;cbData/2 = String length
  StringLeft, Datareceived, CopyOfData, cbData
  XY_eval_and_returnvalue_XYRETURN := Datareceived
}

;*****************************************************************************************************************
